import os
import json

import logging
logger = logging.getLogger(__name__)

__all__ = ['BASE_DIR', 'absolute_path', 'get_int', 'get_bool', 'get_str', 'get_list', 'get_dict']

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def get_int(name, default=0):
    """
    Get a numeric value from environment and convert it accordingly.
    Return default if value does not exist or fails to parse.
    """
    if name not in os.environ:
        return default

    try:
        value = os.environ.get(name, default)
        return int(value)
    except ValueError:
        logger.error('ENV: Non-numeric type found for: {}'.format(name))
        return default


def absolute_path(*args):  # noqa
    return os.path.join(BASE_DIR, *args)


def get_bool(name, default=False):  # noqa
    """
    Get a boolean value from environment variable.
    If the environment variable is not set or value is not one or "true" or
    "false", the default value is returned instead.
    """

    if name not in os.environ:
        return default
    if os.environ[name].lower() in ['true', 'yes', '1', 'y']:
        return True
    elif os.environ[name].lower() in ['false', 'no', '0', 'n']:
        return False
    else:
        return default


def get_str(name, default=None):  # noqa
    """
    Get a string value from environment variable.
    If the environment variable is not set, the default value is returned
    instead.
    """

    value = os.environ.get(name, default)
    if value is None:
        logger.error('ENV: Nothing found for: {}'.format(name))

    return value


def get_list(name, separator=',', default=None):  # noqa
    """
    Get a list of string values from environment variable.
    If the environment variable is not set, the default value is returned
    instead.
    """
    if name not in os.environ:
        if default is None:
            logger.error('ENV: Nothing found for: {}'.format(name))
            default = []
        return default
    return os.environ[name].split(separator)


def get_dict(name, default=None):
    """
    Get JSON encoded string from environment variable and return
    the default if it does not exist.
    """
    if name not in os.environ:
        if default is None:
            logger.error('Nothing found for: {}'.format(name))
            default = {}
        return default
    try:
        dict = json.loads(os.environ[name])
        return dict
    except ValueError:
        logger.error('ENV: Failed to parse value for: {}'.format(name))
        return default



import jwt
from furl import furl
import json
import base64
import requests
import jwcrypto.jwk as jwk

from django.apps import apps
from django.contrib import auth as django_auth
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.core.exceptions import MultipleObjectsReturned
from django.shortcuts import redirect
from django.urls import reverse
from rest_framework.authentication import BaseAuthentication
from rest_framework import exceptions

from dbmi_client.settings import dbmi_settings
from dbmi_client import authz

# Get the app logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)

# Set a key to cache JWKs under in the DBMI.AUTH0 settings
CACHED_JWKS_KEY = '__DBMI_CLIENT_CACHED_JWKS__'


def login(request):
    """
    Builds and returns a redirect response to allow the user to login and send them to the
    supplied URL when successfully logged in. If not defined in LOGIN_REDIRECT_URL, the
    current URL in the request will be used, so the user will return to the original location.
    :param request: The original request object
    :return: The redirect response
    """
    return login_redirect(request)


def login_redirect(request, next_url=None):
    """
    Builds and returns a redirect response to allow the user to login and send them to the
    supplied URL when successfully logged in. If next_url is not passed, and
    it is not defined in LOGIN_REDIRECT_URL, the current URL in the request will be used,
    so the user will return to the original location.
    :param request: The original request object
    :param next_url: The URL users will be sent after login
    :return: Response
    """

    # Ensure the request is cleared of user state
    django_auth.logout(request)

    # Get the url
    login_url = login_redirect_url(request, next_url)

    # Just process the logout and redirect them
    response = redirect(login_url)

    # Do needed logout functions and return the modified response
    return response


def login_redirect_url(request, next_url=None):
    """
    Builds and returns a URL that sends the user to the login, and returns them to the
    supplied URL when successfully logged in. If next_url is not passed, the current URL in the
    request will be used, so the user will return to the original location.
    :param request: The original request object
    :param next_url: The URL users will be sent after login
    :return: Response
    """
    # Check for local login enabled
    if apps.is_installed('dbmi_client.login'):

        # Use local login URL
        login_url = furl(request.build_absolute_uri(reverse('dbmi_login:login')))

    else:

        # Build the URL using DBMI-AuthN
        login_url = furl(dbmi_settings.AUTHN_URL)
        login_url.path.segments.extend(['login', 'auth'])

    # If no next URL, determine where to dump them after logout
    if not next_url:
        if dbmi_settings.LOGIN_REDIRECT_URL:
            next_url = request.build_absolute_uri(dbmi_settings.LOGIN_REDIRECT_URL)

        else:
            next_url = request.build_absolute_uri()

    # Add next url
    logger.debug('Login next URL: {}'.format(next_url))
    login_url.query.params.add(dbmi_settings.LOGIN_REDIRECT_KEY, next_url)

    # Add the default client ID, if specified
    if hasattr(dbmi_settings, 'AUTH0_CLIENT_ID') and getattr(dbmi_settings, 'AUTH0_CLIENT_ID'):
        login_url.query.params.add('client_id', dbmi_settings.AUTH0_CLIENT_ID)

    # Check for branding
    if dbmi_settings.AUTHN_TITLE or dbmi_settings.AUTHN_ICON_URL:

        # Add the included parameters
        branding = {}
        if dbmi_settings.AUTHN_TITLE:
            branding['title'] = dbmi_settings.AUTHN_TITLE

        if dbmi_settings.AUTHN_TITLE:
            branding['icon_url'] = dbmi_settings.AUTHN_ICON_URL

        # Encode it and pass it along
        branding_param = base64.urlsafe_b64encode(json.dumps(branding).encode('utf-8')).decode('utf-8')
        login_url.query.params.add('branding', branding_param)

    logger.debug('Login URL: {}'.format(login_url.url))
    return login_url.url


def logout(request):
    """
    This will prepare the redirect to log the user out (either internally or at DBMI-AuthN)
    :param request: The Django request object
    :return: The response object that takes the user to the logout endpoint
    """
    # Call logout redirect
    return logout_redirect(request)


def logout_redirect(request, next_url=None):
    """
    This is just an alias for `logout` to support older clients
    """
    # Ensure the request is cleared of user state
    django_auth.logout(request)

    # Build the logout URL
    logout_url = logout_redirect_url(request, next_url)

    # Just process the logout and redirect them
    response = redirect(logout_url)

    # Do needed logout functions and return the modified response
    return response


def logout_redirect_url(request, next_url=None):
    """
    This will prepare the redirect URL to log the user out (either internally or at DBMI-AuthN)
    :param request: The Django request object
    :param next_url: A URL that the user should be sent to should they log back in
    :return: The response object that takes the user to the logout endpoint
    """
    # Check for local login enabled
    if apps.is_installed('dbmi_client.login'):

        # Build the URL to DBMI-Client's logout page
        logout_url = furl(request.build_absolute_uri(reverse('dbmi_login:logout')))

    else:

        # Build the URL to DBMI-AuthN's logout endpoint
        logout_url = furl(dbmi_settings.AUTHN_URL)
        logout_url.path.segments.extend(['logout'])

    # If no next URL, determine where to dump them after logout
    if not next_url:
        if dbmi_settings.LOGOUT_REDIRECT_URL:
            next_url = request.build_absolute_uri(dbmi_settings.LOGOUT_REDIRECT_URL)

        else:
            next_url = request.build_absolute_uri()

    logger.debug('Logout next URL: {}'.format(next_url))

    # Add next url
    logout_url.query.params.add(dbmi_settings.LOGOUT_REDIRECT_KEY, next_url)

    # Add the default client ID, if specified
    if hasattr(dbmi_settings, 'AUTH0_CLIENT_ID') and getattr(dbmi_settings, 'AUTH0_CLIENT_ID'):
        logout_url.query.params.add('client_id', dbmi_settings.AUTH0_CLIENT_ID)

    # Check for branding
    if dbmi_settings.AUTHN_TITLE or dbmi_settings.AUTHN_ICON_URL:

        # Add the included parameters
        branding = {}
        if dbmi_settings.AUTHN_TITLE:
            branding['title'] = dbmi_settings.AUTHN_TITLE

        if dbmi_settings.AUTHN_TITLE:
            branding['icon_url'] = dbmi_settings.AUTHN_ICON_URL

        # Encode it and pass it along
        branding_param = base64.urlsafe_b64encode(json.dumps(branding).encode('utf-8')).decode('utf-8')
        logout_url.query.params.add('branding', branding_param)

    logger.debug('Logout URL: {}'.format(logout_url.url))
    return logout_url.url


def dbmi_http_headers(request, content_type='application/json', **kwargs):
    """
    Returns headers to be used for API calls to DBMI services in order to authenticate the caller
    :param request: The Django request
    :param content_type: The content type for the call
    :return: dict
    """

    # Get the JWT
    token = get_jwt(request)

    # Return headers
    headers = {'Authorization': '{}{}'.format(dbmi_settings.JWT_HTTP_PREFIX, token), 'Content-Type': content_type}

    # Add any additional parameters
    headers.update(kwargs)

    return headers


def get_jwt(request):
    """
    Takes a Django request and pulls the JWT from either cookies or HTTP headers
    :param request: The Django request
    :return: The JWT, if found
    """
    # Get the JWT token depending on request type
    if hasattr(request, 'COOKIES') and request.COOKIES.get(dbmi_settings.JWT_COOKIE_NAME):
        return request.COOKIES.get(dbmi_settings.JWT_COOKIE_NAME)

    # Check if JWT in HTTP Authorization header
    elif hasattr(request, 'META') and request.META.get('HTTP_AUTHORIZATION') and dbmi_settings.JWT_HTTP_PREFIX \
            in request.META.get('HTTP_AUTHORIZATION'):

        # Remove prefix and return the token
        return request.META.get('HTTP_AUTHORIZATION').replace(dbmi_settings.JWT_HTTP_PREFIX, '')

    return None


def get_jwt_payload(request, verify=True):

    # Get the JWT token depending on request type
    token = get_jwt(request)

    # Get the payload email
    if not verify:
        return jwt.decode(token, verify=False)

    else:
        return validate_rs256_jwt(token)


def get_jwt_value(request, key, verify=True):

    # Get the payload from above
    payload = get_jwt_payload(request, verify)
    if not payload:
        logger.debug('JWT is invalid, cannot fetch values')
        return None

    return payload.get(key)


def get_jwt_username(request, verify=True):
    return get_jwt_value(request, 'sub', verify)


def get_jwt_email(request, verify=True):
    return get_jwt_value(request, 'email', verify)


def validate_request(request):
    '''
    Pulls the current cookie and verifies the JWT and
    then returns the JWT payload. Returns None
    if the JWT is invalid or missing.
    :param request: The Django request object
    :return: dict
    '''

    # Extract JWT token into a string.
    jwt_string = get_jwt(request)
    algorithm = None

    # Ensure JWT exists
    if not jwt_string:
        return None

    # Determine which Auth0 Client ID (aud) this JWT pertains to.
    try:
        unverified_header = jwt.get_unverified_header(str(jwt_string))
        algorithm = unverified_header.get('alg', 'rs256').lower()

        # Check algorithm
        if algorithm == 'hs256':
            return validate_hs256_jwt(jwt_string)

        elif algorithm == 'rs256':
            return validate_rs256_jwt(jwt_string)

        else:
            logger.error(f'Unsupported JWT algorithm: {algorithm}')
            return None

    except Exception as e:
        logger.exception('Validate error: {}'.format(e), exc_info=True,
                         extra={'algorithm': algorithm})

    return None


def get_public_keys_from_auth0(tenant, refresh=False):
    '''
    Retrieves the public key from Auth0 to verify JWTs. Will
    cache the JSON response from Auth0 in Django settings
    until instructed to refresh the JWKS.
    :param tenant: The Auth0 tenant to fetch JWKs for
    :param refresh: Purges cached JWK and fetches from remote
    :return: dict
    '''

    # If refresh, delete cached key
    if refresh:
        logger.debug('Refresh requested, deleting cached JWKs')
        delattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}')

    try:
        # Look in settings
        if hasattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}'):

            # Parse the cached dict and return it
            return json.loads(getattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}'))

        else:

            logger.debug('Fetching remote JWKS')

            # Build the JWKs URL
            url = furl().set(scheme='https', host='{}.auth0.com'.format(tenant))
            url.path.segments.extend(['.well-known', 'jwks.json'])

            # Make the request
            response = requests.get(url.url)
            response.raise_for_status()

            # Parse it
            jwks = response.json()

            # Cache it
            setattr(dbmi_settings, f'{CACHED_JWKS_KEY}{tenant}', json.dumps(jwks))

            return jwks

    except KeyError as e:
        logger.exception(f'Error getting public keys: {e}', exc_info=True, extra={'tenant': tenant})

    except json.JSONDecodeError as e:
        logger.exception(f'Error getting public keys: {e}', exc_info=True, extra={'tenant': tenant})

    except requests.HTTPError as e:
        logger.exception(f'Error getting public keys: {e}', exc_info=True, extra={'tenant': tenant})

    return None


def retrieve_public_key(tenant, jwt_string):
    '''
    Gets the public key used to sign the JWT from the public JWK
    hosted on Auth0. Auth0 typically only returns one public key
    in the JWK set but to handle situations in which signing keys
    are being rotated, this method is build to search through
    multiple JWK that could be in the set.

    As JWKS are being cached, if a JWK cannot be found, cached
    JWKS is purged and a new JWKS is fetched from Auth0. The
    fresh JWKS is then searched again for the needed key.

    Returns the key ID if found, otherwise returns None
    :param tenant: The Auth0 tenant to fetch JWKs for
    :param jwt_string: The JWT token as a string
    :return: str
    '''

    try:
        # Get the JWK
        jwks = get_public_keys_from_auth0(tenant, refresh=False)
        if not jwks:
            logger.debug('Could not fetch JWKs from Auth0, just fail out now')
            return None

        # Decode the JWTs header component
        unverified_header = jwt.get_unverified_header(str(jwt_string))

        # Check the JWK for the key the JWT was signed with
        rsa_key = get_rsa_from_jwks(jwks, unverified_header['kid'])
        if not rsa_key:
            logger.debug('No matching key found in JWKS, refreshing')
            logger.debug('Unverified JWT key id: {}'.format(unverified_header['kid']))
            logger.debug('Cached JWK keys: {}'.format([jwk['kid'] for jwk in jwks['keys']]))

            # No match found, refresh the jwks
            jwks = get_public_keys_from_auth0(tenant, refresh=True)
            logger.debug('Refreshed JWK keys: {}'.format([jwk['kid'] for jwk in jwks['keys']]))

            # Try it again
            rsa_key = get_rsa_from_jwks(jwks, unverified_header['kid'])
            if not rsa_key:
                logger.warning('Invalid JWT attempt', extra={'unverified_kid': unverified_header['kid']})
                return None

        return rsa_key

    except KeyError as e:
        logger.exception('Error retrieving public keys: {}'.format(e), exc_info=True, extra={'tenant': tenant})

    return None


def get_rsa_from_jwks(jwks, jwt_kid):
    '''
    Searches the JWKS for the signing key used
    for the JWT. Returns a dict of the JWK
    properties if found, None otherwise.
    :param jwks: The set of JWKs from Auth0
    :param jwt_kid: The key ID of the signing key
    :return: dict
    '''
    # Build the dict containing rsa values
    for key in jwks["keys"]:
        if key["kid"] == jwt_kid:
            rsa_key = {
                "kty": key["kty"],
                "kid": key["kid"],
                "use": key["use"],
                "n": key["n"],
                "e": key["e"]
            }

            return rsa_key

    # No matching key found, must refresh JWT keys
    return None


def validate_rs256_jwt(jwt_string):
    '''
    Verifies the given RS256 JWT. Returns the payload
    if verified, otherwise returns None.
    :param jwt_string: JWT as a string
    :return: dict
    '''
    # We need the public key and the client ID to match against
    jwk_pub_key = None
    jwt_client_id = None

    # Determine which Auth0 Client ID (aud) this JWT pertains to.
    try:
        jwt_client_id = str(jwt.decode(jwt_string, verify=False)['aud'])

        # Check if multiple clients are specified at the client level
        if hasattr(dbmi_settings, 'AUTH0_CLIENTS') and getattr(dbmi_settings, 'AUTH0_CLIENTS'):

            # Search client IDs
            tenant = dbmi_settings.AUTH0_CLIENTS.get(jwt_client_id, {}).get('tenant')
            if tenant:
                logger.debug(f'JWT Client ID matched to tenant: {tenant}')

                # Get the public key
                jwk_pub_key = retrieve_public_key(tenant, jwt_string)

            # Log if not found
            else:
                logger.info(f'JWT client {jwt_client_id} could not be matched to any clients')

        # Check for multiple clients at the tenant level
        if not jwk_pub_key and hasattr(dbmi_settings, 'AUTH0_TENANTS') and getattr(dbmi_settings, 'AUTH0_TENANTS'):

            # Try each one
            for tenant in dbmi_settings.AUTH0_TENANTS:
                try:
                    # Get the public key
                    jwk_pub_key = retrieve_public_key(tenant, jwt_string)

                    # A public key was returned, we've got a JWT from this tenant
                    if jwk_pub_key:
                        break

                except PermissionDenied:
                    pass
            else:
                logger.info(f'JWT client {jwt_client_id} could not be matched to '
                            f'any tenants: {dbmi_settings.AUTH0_TENANTS}')

        # If not already matched, try default client
        if not jwk_pub_key and jwt_client_id == dbmi_settings.AUTH0_CLIENT_ID:

            # Get the public key
            jwk_pub_key = retrieve_public_key(dbmi_settings.AUTH0_TENANT, jwt_string)

        if not jwk_pub_key:
            logger.error(f'JWT Client ID could not be matched: {jwt_client_id}')
            return None

    except Exception as e:
        logger.exception(f'Failed to get the aud from jwt payload: {e}', exc_info=True,
                         extra={'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key})
        return None

    # Attempt to validate with each one
    if jwk_pub_key:

        # Get the JWK
        jwk_key = jwk.JWK(**jwk_pub_key)

        # Attempt to validate the JWT (Checks both expiry and signature)
        try:
            payload = jwt.decode(jwt_string,
                                 jwk_key.export_to_pem(private_key=False),
                                 algorithms=['RS256'],
                                 leeway=120,
                                 audience=jwt_client_id)

            return payload

        except jwt.ExpiredSignatureError as e:
            logger.debug("JWT Expired: {}".format(e), extra={
                'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key
            })

        except jwt.InvalidTokenError as e:
            logger.info("Invalid JWT Token: {}".format(e), extra={
                'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key
            })

        except Exception as e:
            logger.exception(f'Error validating JWT: {e}', exc_info=True, extra={
                'jwt_client_id': jwt_client_id, 'jwk_pub_key': jwk_pub_key
            })

    return None


def validate_hs256_jwt(jwt_string):
    '''
    Verifies the given HS256 JWT. Returns the payload
    if verified, otherwise returns None.
    :param jwt_string: JWT as a string
    :return: dict
    '''

    # Check settings for proper setup
    if not dbmi_settings.AUTH0_SECRET or not dbmi_settings.AUTH0_CLIENT_ID:
        logger.error('Cannot verify HS256 tokens without client ID and client secret')
        raise PermissionDenied

    # Determine which Auth0 Client ID (aud) this JWT pertains to.
    jwt_client_id = None
    try:
        jwt_client_id = str(jwt.decode(jwt_string, verify=False)['aud'])

        # Perform the validation
        payload = jwt.decode(jwt_string,
                             base64.b64decode(dbmi_settings.AUTH0_SECRET, '-_'),
                             algorithms=['HS256'],
                             leeway=120,
                             audience=jwt_client_id)

        return payload

    except jwt.ExpiredSignatureError as e:
        logger.debug("JWT Expired: {}".format(e), extra={
            'jwt_client_id': jwt_client_id,
        })

    except jwt.InvalidTokenError as e:
        logger.info("Invalid JWT Token: {}".format(e), extra={
            'jwt_client_id': jwt_client_id,
        })

    except Exception as e:
        logger.exception(f'Error validating JWT: {e}', exc_info=True, extra={
            'jwt_client_id': jwt_client_id,
        })

    return None

###################################################################
#
# Django Custom Authentication Backends
#
###################################################################


class DBMIAuthenticationBackend(object):

    def authenticate(self, request, token=None):
        """
        All versions of this backend follow the same flow:
        1. Check JWT is present
        2. Check JWT validity
        3. Check JWT payload for required user properties
        4. Return existing user or create new one
        5. Sync user, if already exists, with JWT payload
        """
        # Get the token
        if not token:
            token = get_jwt(request)
            if not token:
                return None

        # Validate request
        payload = validate_rs256_jwt(token)
        if not payload:
            return None

        # Get their email and check for their record
        email = payload.get('email')
        username = payload.get('sub')
        if not email or not username:
            logger.error('No sub or email in valid JWT: {}'.format(payload))
            return None

        # The JWT is valid, now get the user object to attach to the request
        user = self._get_user_object(request)
        if not user:

            # Create the user, which will also sync
            user = self._create_user(request)

        else:

            # Sync the user to ensure we've got updated properties
            self._sync_user(request, user)

        return user

    def get_user(self, user_id):
        # Should be implemented by subclass depending on data source for user
        raise SystemError('This method should not be called')

    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user. If model is enabled, this will be an instance of User,
        otherwise an instance of DBMIUser
        """
        # Should be implemented by subclass depending on data source for user
        raise SystemError('This method should not be called')

    def _create_user(self, request):
        """
        Called when a user in the model does not exist. This creates the user in the
        model.
        """
        # Should be implemented by subclass depending on data source for user
        raise SystemError('This method should not be called')

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        try:
            # Get the unverified payload
            payload = get_jwt_payload(request, verify=False)

            # Get properties
            username = payload['sub'].lower()
            email = payload['email'].lower()

            # Check if email or username missing
            if not user.username.lower() == username:
                logger.debug('User\'s username did not match JWT: {} -> {}'.format(user.username, username))
                user.username = username

            if not user.email.lower() == email:
                logger.debug('User\'s email did not match JWT: {} -> {}'.format(user.email, email))
                user.email = email

        except (KeyError, IndexError, TypeError) as e:
            logger.exception('User syncing error: {}'.format(e), exc_info=True,
                             extra={'user': user.id, 'request': request})


class DBMIJWTAuthenticationBackend(DBMIAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    Users objects are an instance of DBMIJWTUser and mimic the properties and methods of Django's built-in
    contrib.auth.models.User model, but with no persistence. All properties will be valid but any attempt to
    save or link these instances to another model instance will fail.
    """
    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user.
        """
        # Create an instance of the JWT user
        user = DBMIJWTUser(request)

        # Sync their profile and return them
        self._sync_user(request, user)

        return user

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        # The user object is built from the JWT on every request so syncing is redundant and
        # not required.
        pass


class DBMIJWTAdminAuthenticationBackend(DBMIAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    Users objects are an instance of DBMIJWTUser and mimic the properties and methods of Django's built-in
    contrib.auth.models.User model, but with no persistence. All properties will be valid but any attempt to
    save or link these instances to another model instance will fail. DBMI AuthZ is consulted for staff/superuser
    access and the User object is prepared as such.
    """
    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user.
        """
        # Create an instance of the JWT user
        user = DBMIJWTUser(request)

        # Sync their profile and return them
        self._sync_user(request, user)

        return user

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        # All sync admin/superuser status
        try:
            # Check if admin/superuser
            if authz.is_admin(request, user.email):
                user.is_staff = True
                user.is_superuser = True
            else:
                user.is_staff = False
                user.is_superuser = False

        except (KeyError, IndexError, TypeError) as e:
            logger.exception('User syncing error: {}'.format(e), exc_info=True,
                             extra={'user': user.id, 'request': request})


class DBMIModelAuthenticationBackend(DBMIAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    If enabled, users are auto-created upon first attempted login with a valid JWT. The
    User model is keyed by the username and email contained in the JWT. Profile and groups are synced
    from the JWT upon each login.
    """

    def user_can_authenticate(self, user):
        """
        Reject users with is_active=False. Custom user models that don't have
        that attribute are allowed.
        """
        is_active = getattr(user, 'is_active', None)
        return is_active or is_active is None

    def get_user(self, user_id):
        UserModel = django_auth.get_user_model()
        try:
            user = UserModel._default_manager.get(pk=user_id)

        except UserModel.DoesNotExist:
            return None

        return user if self.user_can_authenticate(user) else None

    def _get_user_object(self, request):
        """
        Accepts details from the JWT user and returns an object representing
        the request's user. If model is enabled, this will be an instance of User,
        otherwise an instance of DBMIUser
        """
        # Get username and email
        username = get_jwt_username(request, verify=False)
        email = get_jwt_email(request, verify=False)

        # Fetch the current User model
        UserModel = django_auth.get_user_model()
        try:
            # Find the user
            user = UserModel.objects.get(Q(username=username) | Q(email=email))
            logger.debug('Found user: {} : {}'.format(username, email))

            return user

        except MultipleObjectsReturned:
            logger.error('Duplicate users exist for {} : {}'.format(username, email))

        except UserModel.DoesNotExist:
            logger.debug('User does not yet exist: {} : {}'.format(username, email))

        return None

    def _create_user(self, request):
        """
        Called when a user in the model does not exist. This creates the user in the
        model.
        """
        # Check if autocreate is enabled
        if not dbmi_settings.USER_MODEL_AUTOCREATE:
            logger.debug('User autocreate is disabled, boot the current user')
            raise PermissionDenied

        # Get username and email
        username = get_jwt_username(request, verify=False)
        email = get_jwt_email(request, verify=False)

        # Create them
        UserModel = django_auth.get_user_model()
        user = UserModel(username=username, email=email)
        user.set_unusable_password()
        logger.debug('Created user: {}:{}'.format(username, email))

        # Sync them up
        self._sync_user(request, user)

        return user

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object.
        """
        try:
            # Get the unverified payload
            payload = get_jwt_payload(request, verify=False)

            # Get properties
            username = payload['sub'].lower()
            email = payload['email'].lower()

            # Check if email or username missing
            if not user.username.lower() == username:
                logger.debug('User\'s username did not match JWT: {} -> {}'.format(user.username, username))
                user.username = username

            if not user.email.lower() == email:
                logger.debug('User\'s email did not match JWT: {} -> {}'.format(user.email, email))
                user.email = email

            # Save
            user.save()

        except (KeyError, IndexError, TypeError) as e:
            logger.exception('User syncing error: {}'.format(e), exc_info=True,
                             extra={'user': user.id, 'request': request})


class DBMIUsersModelAuthenticationBackend(DBMIModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies).
    Use this authentication backend for sites that are accessible to all users as well as administrator users.
    User model is keyed by the username and email contained in the JWT. `is_staff` and `is_superuser` flags
    are automatically set and synced on appropriate users. Every active user with admin authorization on this site
    will have complete access to everything.
    """

    def _sync_user(self, request, user):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object. Set staff and superuser flags
        if authorizations are valid.
        """
        # Do normal sync first
        super(DBMIModelAuthenticationBackend, self)._sync_user(request, user)

        # Check if admin
        is_admin = authz.is_admin(request, user.email)
        if is_admin:
            logger.debug(f'User: {user.email} has been granted admin/superuser privileges')

        # Ensure the model is updated
        user.is_staff = is_admin
        user.is_superuser = is_admin
        user.save()


class DBMIAdminModelAuthenticationBackend(DBMIModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies) as
    well as admin authorization, either through JWT claims or as a permission in the DBMI AuthZ service.
    Use this authentication backend for sites that are only accessible to admins and no other users.
    User model is keyed by the username and email contained in the JWT. Profile and groups are synced
    from the JWT upon each login.
    """

    def _create_user(self, request):
        """
        This middleware performs exactly like its superclass, with the exception of checking
        an authenticated user's authorizations before creating them in the model. This would
        be used for sites where only admins/superusers/staff should have access.
        """
        # Get username and email
        username = get_jwt_username(request, verify=False)
        email = get_jwt_email(request, verify=False)

        # Before we create a user, we must ensure they have admin authorizations
        if not authz.is_admin(request, email):
            raise PermissionDenied

        # Create them
        UserModel = django_auth.get_user_model()
        user = UserModel(username=username, email=email)
        user.set_unusable_password()
        logger.debug('Created user: {}:{}'.format(username, email))

        # Sync them up
        self._sync_user(request, user, is_admin=True)

        return user

    def _sync_user(self, request, user, is_admin=None):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object. This method is extended to
        add the flag determining admin authorization to prevent multiple
        requests to the authz server.
        """
        # Do normal sync first
        super(DBMIAdminModelAuthenticationBackend, self)._sync_user(request, user)


class DBMISuperuserModelAuthenticationBackend(DBMIAdminModelAuthenticationBackend):

    """
    Clients must have a valid JWT in the request (either in HTTP Authorization headers or in cookies) as
    well as admin authorization, either through JWT claims or as a permission in the DBMI AuthZ service.
    Use this authentication backend for sites that are only accessible to admins and no other users.
    User model is keyed by the username and email contained in the JWT. `is_staff` and `is_superuser` flags
    are automatically set and synced on users. Every active user with authorization on this site
    will have complete access to everything.
    """

    def _sync_user(self, request, user, is_admin=None):
        """
        Called after a user is fetched/created and syncs any additional properties
        from the JWT's payload to the user object. Set staff and superuser flags
        if authorizations are valid.
        """
        # Do normal sync first
        super(DBMIAdminModelAuthenticationBackend, self)._sync_user(request, user)

        # Check if admin
        if is_admin is None:
            is_admin = authz.is_admin(request, user.email)

        # Ensure the model is updated
        user.is_staff = is_admin
        user.is_superuser = is_admin
        user.save()

        # If not admin (indicates they used to be), save and raise exception
        if not is_admin:
            logger.debug('User was superuser, but is now missing authz, booting them: {}'.format(user.username))
            raise PermissionDenied

###################################################################
#
# Django user object for user-less databases
#
###################################################################


class DBMIJWTUser(AnonymousUser):
    """
    This class represents a JWT user for an application that does not persist those
    users to the store. It provides a request.user object to access user properties
    but will throw exceptions if ever used in the context of database operations.
    """
    email = None
    username = None
    id = None
    pk = None
    is_active = True
    is_staff = False
    is_superuser = False

    def __init__(self, request):

        # Get the payload
        payload = get_jwt_payload(request, verify=False)

        # Set properties
        self.username = payload.get('sub')
        self.id = payload.get('sub')
        self.email = payload.get('email')

    def __str__(self):
        return self.id

    def __eq__(self, other):
        return hasattr(other, 'id') and self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def save(self):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    def delete(self):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    def set_password(self, raw_password):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    def check_password(self, raw_password):
        raise NotImplementedError("Django doesn't provide a DB representation for DBMIJWTUser.")

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    def get_username(self):
        return self.username

    def get_group_permissions(self, obj=None):
        """
        Return a list of permission strings that this user has through their
        groups. Query all available auth backends. If an object is passed in,
        return only permissions matching this object.
        """
        permissions = set()
        return permissions

    def get_all_permissions(self, obj=None):
        return _user_get_all_permissions(self, obj)

    def has_perm(self, perm, obj=None):
        """
        Return True if the user has the specified permission. Query all
        available auth backends, but return immediately if any backend returns
        True. Thus, a user who has permission from a single auth backend is
        assumed to have permission in general. If an object is provided, check
        permissions for that object.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        # Otherwise we need to check the backends.
        return _user_has_perm(self, perm, obj)

    def has_perms(self, perm_list, obj=None):
        """
        Return True if the user has each of the specified permissions. If
        object is passed, check if the user has all required perms for it.
        """
        return all(self.has_perm(perm, obj) for perm in perm_list)

    def has_module_perms(self, app_label):
        """
        Return True if the user has any permissions in the given app label.
        Use similar logic as has_perm(), above.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        return _user_has_module_perms(self, app_label)


# A few helper functions for common logic between User and AnonymousUser.
def _user_get_all_permissions(user, obj):
    permissions = set()

    # TODO: Query DBMI AuthZ and return all permission items 'item.permission'

    return permissions


def _user_has_perm(user, perm, obj):
    """
    A backend can raise `PermissionDenied` to short-circuit permission checking.
    """
    # Get all perms and compare
    try:
        for permission in _user_get_all_permissions(user, obj):
            if permission == '{}.{}'.format(obj, perm):
                return True
    except PermissionDenied:
        return False

    return False


def _user_has_module_perms(user, app_label):
    """
    A backend can raise `PermissionDenied` to short-circuit permission checking.
    """
    # Get all perms and compare
    try:
        for perm in _user_get_all_permissions(user, app_label):
            if '{}.{}'.format(app_label, perm) == perm:
                return True
    except PermissionDenied:
        return False

    return False

###################################################################
#
# Django Rest Framework (DRF) Custom Authentication/Authorization
#
###################################################################


class DBMIUser(BaseAuthentication):
    """
    Authentication method for DBMI API methods
    """
    def authenticate(self, request):

        # Get the JWT
        token = get_jwt(request)
        if not token:
            raise exceptions.NotAuthenticated

        # User has a valid JWT from SciAuth
        payload = validate_rs256_jwt(token)
        if not payload:
            raise exceptions.AuthenticationFailed

        # Return the user's email to attach to the request object (request.user)
        # Also, return the authz dictionary contained in the JWT claims, if present (request.auth)
        return payload.get('email'), payload.get(dbmi_settings.JWT_AUTHZ_NAMESPACE)


class DBMIModelUser(BaseAuthentication):
    """
    Authentication method for DBMI API methods
    """
    def authenticate(self, request):

        # Get the JWT
        token = get_jwt(request)
        if not token:
            raise exceptions.NotAuthenticated

        # Call the standard Django authenticate method, that will in
        # turn call DBMIModelAuthenticationBackend.authenticate
        user = django_auth.authenticate(request, token=token)
        if not user:
            raise exceptions.AuthenticationFailed

        # Check if JWT contains AuthZ
        auth = None
        if dbmi_settings.JWT_AUTHZ_NAMESPACE:

            # User has a valid JWT from SciAuth
            auth = get_jwt_payload(request, verify=False).get('JWT_AUTHZ_NAMESPACE')

        # Return the user's email to attach to the request object (request.user)
        # Also, return the authz dictionary contained in the JWT claims, if present (request.auth)
        return user, auth



from raven.contrib.django.client import DjangoClient
from raven.contrib.django.client import get_client_ip

from dbmi_client.authn import get_jwt_username, get_jwt_email, get_jwt_value


class DBMISentryClient(DjangoClient):
    """
    This client class prevents issues during authentication where
    Sentry would try to access auth state if an error was logged
    during authentication, leading to an infinite recursion issue.
    Instead of relying on the request.user object for user properties,
    we pull them straight from the JWT, if available.
    """

    def get_user_info(self, request):

        # Get info from token header
        return {
            'ip_address': get_client_ip(request.META),
            'username': get_jwt_username(request, verify=False),
            'email': get_jwt_email(request, verify=False),
            'client': get_jwt_value(request, 'aud', verify=False),
            'tenant': get_jwt_value(request, 'iss', verify=False),
        }



from furl import furl
import requests
import json

from dbmi_client import authn
from dbmi_client.settings import dbmi_settings


# Get the app logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


def create_dbmi_user(request, **profile):
    logger.debug("Creating DBMI user")

    # Get the JWT
    email = authn.get_jwt_email(request, verify=False)

    # Update kwargs
    profile['email'] = email

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(['api', 'register', ''])

    response = requests.post(url.url, headers=authn.dbmi_http_headers(request), data=json.dumps(profile))
    if not response.ok:
        logger.error('Create user response: {}'.format(response.content))

    return response


def send_email_confirmation(request, success_url):
    logger.debug("Sending confirmation email")

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(['api', 'register', 'send_confirmation_email', ''])

    data = {
        'success_url': success_url
    }

    # Make the call
    response = requests.post(url.url, headers=authn.dbmi_http_headers(request), data=json.dumps(data))
    if not response.ok:
        logger.error('Confirmation email response: {}'.format(response.content))

    return response.ok


def check_email_confirmation(request):
    logger.debug("Checking email confirmation")

    # Build the URL (needs trailing slash)
    url = furl(dbmi_settings.REG_URL)
    url.path.segments.extend(['api', 'register', ''])
    url.query.params.add('email', authn.get_jwt_email(request, verify=False))

    # Make the call
    response = requests.get(url.url, headers=authn.dbmi_http_headers(request))
    if not response.ok:
        logger.error('Confirmation email response: {}'.format(response.content))
        return None

    try:
        # Parse the profile for the status
        email_status = response.json()['results'][0]['email_confirmed']
        logger.debug('Email confirmation status: {}'.format(email_status))

        return email_status

    except (KeyError, IndexError) as e:
        logger.error('Failed parsing profile: {}'.format(e))

    return None





from django.conf.urls import url

from dbmi_client.login import views

app_name = "dbmi_login"

urlpatterns = [
    url(r'^login/?$', views.login,  name='login'),
    url(r'^callback/?$', views.callback,  name='callback'),
    url(r'^logout/?$', views.login,  name='logout'),
    url(r'^jwt/?$', views.token,  name='jwt'),
    url(r'^$', views.token,  name='jwt'),
]





from enum import Enum

import requests
from furl import furl
from rest_framework import status
from django.conf import settings

from dbmi_client.settings import dbmi_settings
from dbmi_client import authn

import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


# Set the possible permissions
class FileserviceRole(Enum):
    Readers = '{}__READERS'.format(dbmi_settings.FILESERVICE_GROUP)
    Writers = '{}__WRITERS'.format(dbmi_settings.FILESERVICE_GROUP)
    Downloaders = '{}__DOWNLOADERS'.format(dbmi_settings.FILESERVICE_GROUP)
    Uploaders = '{}__UPLOADERS'.format(dbmi_settings.FILESERVICE_GROUP)
    admins = '{}__ADMINS'.format(dbmi_settings.FILESERVICE_GROUP)


def check_groups(request):
    """
    This checks Fileservice to ensure the specified group has been created.
    :param request: The current request object
    :return: bool
    """

    # Build the URL
    url = furl(settings.FILESERVICE_URL)
    url.path.segments.extend(['filemaster', 'groups', dbmi_settings.FILESERVICE_GROUP])

    response = None
    try:
        # Make the request
        response = requests.get(url.url, headers=_headers(request))
        response.raise_for_status()

        # Return whether it exists or not
        groups = response.json()
        for group in groups:
            if settings.FILESERVICE_GROUP in group['name']:
                return True

    except requests.HTTPError as e:
        logger.exception(e)
        logger.error('Group check error: {}'.format(response.content))

    except TypeError as e:
        logger.exception(e)
        logger.error('Group check error: {}'.format(response.content))

    return False


def create_file(request, filename, metadata={}, tags=[]):
    """
    Creates a file in Fileservice. This initializes the process by which a file can be stored through
    Fileservice and track metadata/tags.
    :param request: The original request
    :param filename: The filename to use
    :param metadata: Any metadata to include with the file
    :param tags: A list of tags
    :return: The file's UUID
    """

    # Ensure groups exist.
    if not check_groups(request):
        logger.error('Groups do not exist or failed to create')
        return None

    # Build the request.
    data = {
        'permissions': [
            dbmi_settings.FILESERVICE_GROUP
        ],
        'metadata': metadata,
        'filename': filename,
        'tags': tags,
    }

    # Build the URL
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments.extend(['filemaster', 'api', 'file'])

    response = None
    try:
        # Make the request
        response = requests.post(url.url, headers=_headers(request), data=data)

        # Check for known issues
        if response.status_code == status.HTTP_403_FORBIDDEN:
            logger.error('User is forbidden: {}'.format(response))
            raise FileservicePermissionError(response)

        # Check for unexpected errors
        response.raise_for_status()

        # Parse the response
        file = response.json()

        # Get the UUID.
        uuid = file['uuid']

        return uuid

    except requests.HTTPError as e:
        logger.exception(e)
        logger.error('File creation error: {}'.format(response))


def get_file_post(request, uuid):
    """
    This method generates the data for a presigned POST operation for uploading the file to S3. Take the
    returned url and dict and make a POST with your binary file. This will place the file in S3 in a location
    created by Fileservice.
    :param request:
    :param uuid:
    :return:
    """

    # Form the request for the file link
    params = {
        'cloud': 'aws',
        'bucket': dbmi_settings.FILESERVICE_AWS_BUCKET,
        'expires': 100
    }

    # Build the URL
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments.extend(['filemaster', 'api', 'file', uuid, 'post'])

    response = None
    try:
        # Make the request
        response = requests.get(url.url, headers=_headers(request), params=params)

        # Check for known issues
        if response.status_code == status.HTTP_403_FORBIDDEN:
            logger.error('User is forbidden: {}'.format(response))
            raise FileservicePermissionError(response)

        # Check for unexpected errors
        response.raise_for_status()

        # Parse the response
        return response.json()

    except requests.HTTPError as e:
        logger.exception(e)
        logger.error('Get file post error: {}'.format(response))


def set_file_uploaded(request, uuid):

    # Build the URL
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments.extend(['filemaster', 'api', 'file', uuid, 'uploadcomplete'])

    response = None
    try:
        # Make the request
        response = requests.get(url.url, headers=_headers(request))

        # Check for known issues
        if response.status_code == status.HTTP_403_FORBIDDEN:
            logger.error('User is forbidden: {}'.format(response))
            raise FileservicePermissionError(response)

        # Check for unexpected errors
        response.raise_for_status()

        # Parse the response
        return response.ok

    except requests.HTTPError as e:
        logger.exception(e)
        logger.error('Get file post error: {}'.format(response))


def get_file_download(request, uuid):
    """
    Returns an S3 URL to download the requested file. The URL has a limited lifespan and will expire
    after an hour (TODO: Confirm this)
    :param request:
    :param uuid:
    :return:
    """
    # Build the URL
    url = furl(dbmi_settings.FILESERVICE_URL)
    url.path.segments.extend(['filemaster', 'api', 'file', uuid, 'download'])

    response = None
    try:
        # Make the request
        response = requests.get(url.url, headers=_headers(request))

        # Check for known issues
        if response.status_code == status.HTTP_403_FORBIDDEN:
            logger.error('User is forbidden: {}'.format(response))
            raise FileservicePermissionError(response)

        # Check for unexpected errors
        response.raise_for_status()

        # Parse the response and return the URL
        download = response.json()

        # Parse the response
        return download.get('url')

    except requests.HTTPError as e:
        logger.exception(e)
        logger.error('Get file post error: {}'.format(response))


def _headers(request):
    """
    Returns the headers to use for Fileservice requests. If a Fileservice token is specified in settings,
    this will use that token to sign calls, otherwise, the current user's JWT will be used.
    :param request: The request
    :return: dict
    """
    if dbmi_settings.FILESERVICE_TOKEN:

        # Use the service token from environment
        return {"Authorization": 'Token {}'.format(dbmi_settings.FILESERVICE_TOKEN), 'Content-Type': 'application/json'}

    else:

        # Get the JWT
        token = authn.get_jwt(request)

        # Use the service token from environment
        return {"Authorization": 'JWT {}'.format(token), 'Content-Type': 'application/json'}


# Indicates the requesting user does not have sufficient permissions for the operation
class FileservicePermissionError(Exception):
    pass


# Indicates the current app's group does not exist
class FileserviceGroupError(Exception):
    pass


# Indicates the current app's targeted bucket does not exist in Fileservice
class FileserviceBucketError(Exception):
    pass



import requests
import json
import base64
import furl

from django.shortcuts import render, redirect, reverse
from django.http import HttpResponse, QueryDict
from dbmi_client.settings import dbmi_settings
from dbmi_client.authn import validate_request, get_jwt
from django.contrib import auth as django_auth

# Get the logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


def token(request):

    # Ensure we've got a JWT
    jwt = get_jwt(request)
    if not jwt:
        return redirect('dbmi_login:login')

    # Set the token
    context = {
        'jwt': jwt
    }

    return render(request, template_name='dbmi_client/login/jwt.html', context=context)


def login(request):
    """
    Landing point to force user log in.

    This URL is a catch-all to see if a user is already logged in. The next Querystring should be set to
    redirect if the user is found to be logged in, or after they log in.
    """
    logger.debug("Checking if user is logged in already.")

    # Check for an existing valid DBMI JWT
    if validate_request(request):

        # Determine where to send them
        if hasattr(dbmi_settings, 'LOGIN_REDIRECT_KEY') and request.GET.get(dbmi_settings.LOGIN_REDIRECT_KEY):
            redirect_url = request.GET.get(dbmi_settings.LOGIN_REDIRECT_KEY)

        elif hasattr(dbmi_settings, 'LOGIN_REDIRECT_URL'):
            redirect_url = dbmi_settings.LOGIN_REDIRECT_URL

        else:
            redirect_url = request.build_absolute_uri(reverse('dbmi_login:jwt'))

        # Log their next destination
        logger.debug(f'Logged in, forward user to: {redirect_url}')

        return redirect(redirect_url)

    # Build a URL with the root URI
    callback_url = furl.furl(request.build_absolute_uri(reverse('dbmi_login:callback')))

    # Pass along any parameters as base64 encoded
    query = base64.urlsafe_b64encode(request.META.get('QUERY_STRING').encode('utf-8')).decode('utf-8')
    callback_url.query.params.add('query', query)

    # Initialize the context.
    context = {
        'callback_url': callback_url.url,
        'auth0_client_id': dbmi_settings.AUTH0_CLIENT_ID,
        'auth0_domain': '{}.auth0.com'.format(dbmi_settings.AUTH0_TENANT),
        'title': dbmi_settings.AUTHN_TITLE,
        'icon_url': dbmi_settings.AUTHN_ICON_URL
    }

    return render(request, template_name='dbmi_client/login/login.html', context=context)


def callback(request):
    """
    Callback from Auth0

    This endpoint is called by auth0 with a code that lets us know the user logged into their Identity Provider successfully.
    We need to use the code to gather the user information from Auth0 and establish the DBMI_JWT cookie.
    """
    logger.debug("Call returned from Auth0.")

    # Fetch some of the request parameters
    query = None
    login_url = furl.furl(request.build_absolute_uri(reverse('dbmi_login:login')))
    try:
        # Get the original query sent to dbmiauth
        query = QueryDict(base64.urlsafe_b64decode(request.GET.get('query').encode('utf-8')).decode('utf-8'))

        # Get the return URL
        login_url = login_url.url + '?{}'.format(query.urlencode('/'))

    except Exception as e:
        logger.error('Failed to parse query parameters: {}'.format(e), exc_info=True, extra={'request': request})

    # This is a code passed back from Auth0 that is used to retrieve a token (Which is used to retrieve user info).
    code = request.GET.get('code')
    if not code:
        logger.error('No code from Auth0', exc_info=True, extra={'request': request})

        # Redirect back to the auth screen and attach the original query
        return redirect(login_url)

    json_header = {'content-type': 'application/json'}

    # This is the Auth0 URL we post the code to in order to get token.
    token_url = 'https://%s.auth0.com/oauth/token' % dbmi_settings.AUTH0_TENANT

    # Build a URL with the root URI
    callback_url = furl.furl(request.build_absolute_uri(reverse('dbmi_login:callback')))

    # Information we pass to auth0, helps identify us and our request.
    token_payload = {
        'client_id': dbmi_settings.AUTH0_CLIENT_ID,
        'client_secret': base64.b64decode(dbmi_settings.AUTH0_SECRET.encode()).decode(),
        'redirect_uri': callback_url.url,
        'code': code,
        'grant_type': 'authorization_code'
    }

    # Post the code to get the token from Auth0.
    token_response = requests.post(token_url, data=json.dumps(token_payload), headers=json_header)
    if not token_response.ok:
        logger.error('Failed to exchange token', exc_info=True, extra={
            'request': request, 'response': token_response.content,
            'status': token_response.status_code, 'url': token_url,
        })

        # Redirect back to the auth screen and attach the original query
        return redirect(login_url)

    # Get tokens
    token_info = token_response.json()

    # URL we post the token to get user info.
    url = 'https://%s.auth0.com/userinfo?access_token=%s'
    user_url = url % (dbmi_settings.AUTH0_TENANT, token_info.get('access_token', ''))

    # Get the user info from auth0.
    user_response = requests.get(user_url)
    if not user_response.ok:
        logger.error('Failed to get user info', exc_info=True, extra={
            'request': request, 'response': user_response.content,
            'status': user_response.status_code, 'url': user_url,
        })

        # Redirect back to the auth screen and attach the original query
        return redirect(login_url)

    # Get user info
    user_info = user_response.json()
    email = user_info.get('email')
    jwt = token_info.get('id_token')
    if email and jwt:

        # Redirect the user to the page they originally requested.
        if hasattr(dbmi_settings, 'LOGIN_REDIRECT_KEY') and query.get(dbmi_settings.LOGIN_REDIRECT_KEY):
            redirect_url = query.get(dbmi_settings.LOGIN_REDIRECT_KEY)

        elif hasattr(dbmi_settings, 'LOGIN_REDIRECT_URL'):
            redirect_url = dbmi_settings.LOGIN_REDIRECT_URL

        else:
            redirect_url = request.build_absolute_uri(reverse('dbmi_login:jwt'))

        logger.debug('Redirecting user to: {}'.format(redirect_url))

        # Set the JWT into a cookie in the response.
        response = redirect(redirect_url)
        response.set_cookie(dbmi_settings.JWT_COOKIE_NAME, jwt, domain=dbmi_settings.JWT_COOKIE_DOMAIN, httponly=True)

        return response

    else:
        logger.error("No email/jwt returned for user info, cannot proceed", exc_info=True, extra={'user_info': user_info})
        return HttpResponse(status=500)


def logout(request):
    """
    User logout

    This endpoint logs out the user session from the dbmi_authn Django app.
    """
    # See if they are logged in
    if validate_request(request):

        # Build the logout URL
        url = furl.furl(f'https://{dbmi_settings.AUTH0_TENANT}.auth0.com/v2/logout')

        # Add the client ID
        url.query.params.add('client_id', dbmi_settings.AUTH0_CLIENT_ID)

        # Look for next url
        if request.GET.get(dbmi_settings.LOGOUT_REDIRECT_KEY):

            # Get the passed URL
            next_url = request.GET.get(dbmi_settings.LOGOUT_REDIRECT_KEY)
            logger.debug('Will log user out and redirect to: {}'.format(next_url))

            # Redirect the user
            url.query.params.set('returnTo', next_url)

        else:
            logger.debug('Will log user out and redirect to log out page')

            # Redirect the user to the landing page
            url.query.params.set('returnTo', request.build_absolute_uri(reverse('dbmi_login:logout')))

        # Log the URL
        logger.debug(f'Logout URL: {url.url}')

        # Ensure the request is cleared of user state
        django_auth.logout(request)

        # Create the response
        response = redirect(url.url)

        # Set the URL and purge cookies
        response.delete_cookie(dbmi_settings.JWT_COOKIE_NAME, domain=dbmi_settings.JWT_COOKIE_DOMAIN)

        return response

    else:
        logger.debug('User has been logged out, sending to logout page')

        # Render the logout landing page
        return render(request, 'dbmi_client/login/logout.html')


from django.core.exceptions import PermissionDenied
from django.contrib import auth as django_auth

from dbmi_client.settings import dbmi_settings
from dbmi_client import authn
from dbmi_client import authz

# Get the app logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)


def dbmi_user(view):
    '''
    Decorator to only check if the current user's JWT is valid
    :param function:
    :type function:
    :return:
    :rtype:
    '''
    def wrap(request, *args, **kwargs):

        # Check for current user
        if not request.user or not request.user.is_authenticated:
            return authn.login_redirect(request)

        # Let it go
        return view(request, *args, **kwargs)

    return wrap


def dbmi_admin(view):
    '''
    Decorator to check for valid JWT and admin permissions
    :param function:
    :type function:
    :return:
    :rtype:
    '''
    def wrap(request, *args, **kwargs):

        # Check for current user
        if not request.user or not request.user.is_authenticated:
            return authn.login_redirect(request)

        # Get the payload
        payload = authn.get_jwt_payload(request, verify=False)

        # Check claims in the JWT first, as it is least costly.
        if authz.jwt_has_authz(payload, authz.JWT_AUTHZ_GROUPS, dbmi_settings.AUTHZ_ADMIN_GROUP):
            return view(request, *args, **kwargs)

        # Get their email address
        email = authn.get_jwt_email(request, verify=False)

        # Now consult the AuthZ server
        if authz.has_permission(request, email, dbmi_settings.CLIENT, dbmi_settings.AUTHZ_ADMIN_PERMISSION):
            return view(request, *args, **kwargs)

        # Possibly store these elsewhere for records
        # TODO: Figure out a better way to flag failed access attempts
        logger.warning('{} Failed {} permission on {}'.format(email,
                                                              dbmi_settings.AUTHZ_ADMIN_PERMISSION,
                                                              dbmi_settings.CLIENT))

        raise PermissionDenied

    return wrap


def dbmi_group(group):
    '''
    Decorator that accepts a group and checks the request user to see if they
    belong to said group.
    :param group: The group
    :type group: str
    :return: function
    '''

    def real_decorator(view):

        def wrap(request, *args, **kwargs):

            # Check for current user
            if not request.user or not request.user.is_authenticated:
                return authn.login_redirect(request)

            # Get the payload
            payload = authn.get_jwt_payload(request, verify=False)

            # Check claims in the JWT first, as it is least costly.
            if authz.jwt_has_authz(payload, authz.JWT_AUTHZ_GROUPS, group):
                return view(request, *args, **kwargs)

            # Possibly store these elsewhere for records
            # TODO: Figure out a better way to flag failed access attempts
            logger.warning('{} Failed {} group on {}'.format(payload.get('email'), group, dbmi_settings.CLIENT))

            # Forbid if nothing else
            raise PermissionDenied

        wrap.__doc__ = view.__doc__
        wrap.__name__ = view.__name__
        return wrap

    return real_decorator


def dbmi_role(role):
    '''
    Decorator that accepts an item string that is used to retrieve
    roles from JWT AuthZ claims.
    :param role: The role
    :type role: str
    :return: function
    '''

    def real_decorator(view):

        def wrap(request, *args, **kwargs):

            # Check for current user
            if not request.user or not request.user.is_authenticated:
                return authn.login_redirect(request)

            # Get the payload
            payload = authn.get_jwt_payload(request, verify=False)

            # Check claims in the JWT first, as it is least costly.
            if authz.jwt_has_authz(payload, authz.JWT_AUTHZ_ROLES, role):
                return view(request, *args, **kwargs)

            # Possibly store these elsewhere for records
            # TODO: Figure out a better way to flag failed access attempts
            logger.warning('{} Failed {} group on {}'.format(payload.get('email'), role, dbmi_settings.CLIENT))

            # Forbid if nothing else
            raise PermissionDenied

        wrap.__doc__ = view.__doc__
        wrap.__name__ = view.__name__
        return wrap

    return real_decorator


def dbmi_app_permission(permission):
    '''
    Decorator that accepts an item string that is used to retrieve
    permissions from SciAuthZ.
    :param permission: The permission
    :type permission: str
    :return: function
    '''

    def real_decorator(view):

        def wrap(request, *args, **kwargs):

            # Check for current user
            if not request.user or not request.user.is_authenticated:
                return authn.login_redirect(request)

            # Get the payload
            payload = authn.get_jwt_payload(request, verify=False)

            # Check claims in the JWT first, as it is least costly.
            if authz.jwt_has_authz(payload, authz.JWT_AUTHZ_PERMISSIONS, permission):
                return view(request, *args, **kwargs)

            # Get their email address
            email = authn.get_jwt_email(request, verify=False)

            # Check permission on the app
            if authz.has_permission(request, email, dbmi_settings.CLIENT, permission):
                return view(request, *args, **kwargs)

            # Possibly store these elsewhere for records
            # TODO: Figure out a better way to flag failed access attempts
            logger.warning('{} Failed {} permission on {}'.format(email, permission, dbmi_settings.CLIENT))

            # Forbid if nothing else
            raise PermissionDenied

        wrap.__doc__ = view.__doc__
        wrap.__name__ = view.__name__
        return wrap

    return real_decorator


def dbmi_item_permission(item, permission):
    '''
    Decorator that accepts an item string that is checked for the passed permission. Item is an arbitrary
    string the application uses internally for specific or object-level permissions.
    :param item: The item
    :type item: str
    :param permission: The permission
    :type permission: str
    :return: function
    '''

    def real_decorator(view):

        def wrap(request, *args, **kwargs):

            # Check for current user
            if not request.user or not request.user.is_authenticated:
                return authn.login_redirect(request)

            # Get their email address
            email = authn.get_jwt_email(request, verify=False)

            # Check permission
            if authz.has_permission(request, email, item, permission):
                return view(request, *args, **kwargs)

            # Possibly store these elsewhere for records
            # TODO: Figure out a better way to flag failed access attempts
            logger.warning('{} Failed {} permission on {}'.format(email, permission, dbmi_settings.CLIENT))

            # Forbid if nothing else
            raise PermissionDenied

        wrap.__doc__ = view.__doc__
        wrap.__name__ = view.__name__
        return wrap

    return real_decorator



import sys
import logging


def config(service, sentry=True):

    # Set the standard configurations
    config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'console': {
                'format': f'[DJANGO] - [{service.upper()}] - [%(asctime)s][%(levelname)s]'
                          f'[%(name)s.%(funcName)s:%(lineno)d] - %(message)s',
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'console',
                'stream': sys.stdout,
            },
        },
        'root': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
        'loggers': {
            'django': {
                'handlers': ['console'],
                'level': 'WARNING',
                'propagate': True,
            },
            'django.request': {
                'handlers': ['console'],
                'level': 'ERROR',
                'propagate': True,
            },
        },
    }

    # Add Sentry/Raven
    if sentry:

        # Add the handler
        config['handlers']['sentry'] = {
            'level': 'ERROR',  # To capture more than ERROR, change to WARNING, INFO, etc.
            'class': 'raven.contrib.django.raven_compat.handlers.SentryHandler',
            'tags': {'custom-tag': 'x'},
        }
        config['root']['handlers'].append('sentry')

        # Add the loggers
        config['loggers']['raven'] = {
            'level': 'WARNING',
            'handlers': ['console'],
            'propagate': False,
        }

        config['loggers']['sentry.errors'] = {
            'level': 'WARNING',
            'handlers': ['console'],
            'propagate': False,
        }

    return config


def add_level(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
       raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
       raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
       raise AttributeError('{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)
    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)



from __future__ import absolute_import, unicode_literals

import os
import warnings
import logging

from django.conf import settings
from django.test.signals import setting_changed

from dbmi_client import environment as env

# Always import this module as follows:
# from dbmi_client import settings [as dbmi_settings]

# Don't import directly CONFIG or PANELs, or you will miss changes performed
# with override_settings in tests.

# Set service URLs based on environment
DBMI_ENVIRONMENTS = {
    'prod': {
        'AUTHN_URL': 'https://authentication.dbmi.hms.harvard.edu:8001',
        'AUTHZ_URL': 'https://authorization.dbmi.hms.harvard.edu:8003',
        'REG_URL': 'https://registration.dbmi.hms.harvard.edu:8005',
        'FILESERVICE_URL': 'https://files.dbmi.hms.harvard.edu',
        'JWT_AUTHZ_NAMESPACE': 'https://authorization.dbmi.hms.harvard.edu',
    },
    'dev': {
        'AUTHN_URL': 'https://authentication.aws.dbmi-dev.hms.harvard.edu',
        'AUTHZ_URL': 'https://authorization.aws.dbmi-dev.hms.harvard.edu',
        'REG_URL': 'https://registration.aws.dbmi-dev.hms.harvard.edu',
        'FILESERVICE_URL': 'https://fileservice.aws.dbmi-dev.hms.harvard.edu',
        'JWT_AUTHZ_NAMESPACE': 'https://authorization.aws.dbmi-dev.hms.harvard.edu',
    }
}

CONFIG_DEFAULTS = {

    # The identifier for this service and/or project
    'CLIENT': None,

    # Client options, assume production environment
    'ENVIRONMENT': 'prod',

    # Set prod URLs
    'AUTHN_URL': None,
    'AUTHZ_URL': None,
    'REG_URL': None,

    # Optionally disable logging
    'LOGGER_NAME': 'dbmi_client',
    'ENABLE_LOGGING': True,
    'LOG_LEVEL': logging.WARNING,

    # Universal login screen branding
    'AUTHN_TITLE': 'DBMI Client',
    'AUTHN_ICON_URL': 'https://authentication.dbmi.hms.harvard.edu/static/hms_shield.png',

    # AuthZ groups/roles/permissions
    'AUTHZ_ADMIN_GROUP': 'dbmi-admin',
    'AUTHZ_ADMIN_PERMISSION': 'MANAGE',
    'AUTHZ_USER_GROUP': 'dbmi-user',

    # JWT bits
    'JWT_AUTHZ_NAMESPACE': None,
    'JWT_HTTP_PREFIX': 'JWT ',
    'JWT_COOKIE_NAME': 'DBMI_JWT',
    'JWT_COOKIE_DOMAIN': '.dbmi.hms.harvard.edu',

    # Auth0 account details
    'AUTH0_CLIENT_ID': None,
    'AUTH0_SECRET': None,  # Only needed if login module is enabled
    'AUTH0_TENANT': None,
    'AUTH0_SCOPE': 'openid email',

    # Auth0 clients
    'AUTH0_CLIENTS': None,  # Use this dictionary to specify multiple auth0 tenant: {client ID, secret} configs

    # Auth0 tenants
    'AUTH0_TENANTS': None,  # Use this list to specify Auth0 tenants whose clients should be considered valid here

    # Configurations surrounding usage of a local Django user model
    'USER_MODEL_ENABLED': False,
    'USER_MODEL_AUTOCREATE': True,

    # These configurations are specific to DRF related auth/permissions
    'DRF_OBJECT_OWNER_KEY': 'user',

    # Filservice
    'FILESERVICE_BUCKETS': [],  # The name of the S3 buckets Fileservice should use
    'FILESERVICE_GROUP': None,  # Typically would be the same as CLIENT

    # Login settings
    'LOGIN_REDIRECT_KEY': 'next',  # The query parameter key specifying where logged in users should be sent
    'LOGIN_REDIRECT_URL': None,  # The post-login URL to send users to if not specified by 'LOGIN_REDIRECT_KEY'

    # Logout settings
    'LOGOUT_REDIRECT_KEY': 'next',  # The query parameter key specifying where logged out users should be sent
    'LOGOUT_REDIRECT_URL': None,  # The post-logout URL to send users to if not specified by 'LOGOUT_REDIRECT_KEY'
}

# List of settings that cannot be defaulted and must be user-defined
REQUIRED_SETTINGS = ('CLIENT', 'AUTH0_CLIENT_ID', 'AUTH0_TENANT')

# List of settings that have been removed
REMOVED_SETTINGS = ()


class DBMISettings(object):
    """
    A settings object, that allows DBMI Client settings to be accessed as properties.
    For example:
        from dbmi_client.settings import dbmi_settings
        print(dbmi_settings.AUTHZ_ADMIN_GROUP)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """
    def __init__(self, defaults=None):
        self.defaults = defaults or CONFIG_DEFAULTS
        self._cached_attrs = set()

    @property
    def user_settings(self):

        # Check to see if user configs have been loaded or not
        if not hasattr(self, '_user_settings'):

            try:
                # Load user-specified configurations
                user_settings = getattr(settings, 'DBMI_CLIENT_CONFIG', {})

                # Update the client config with pre-defined environment URLs, etc
                if user_settings.get('ENVIRONMENT') in DBMI_ENVIRONMENTS:
                    user_settings.update(DBMI_ENVIRONMENTS[user_settings.get('ENVIRONMENT')])

                else:
                    # Check for them in environment
                    for key in DBMI_ENVIRONMENTS['prod'].keys():
                        if env.get_str('DBMI_{}'.format(key.upper())):
                            user_settings[key] = env.get_str('DBMI_{}'.format(key.upper()))

                # Check them
                self._user_settings = self.__check_user_settings(user_settings)

            except Exception as e:
                raise SystemError('DBMI Client settings are invalid: {}'.format(e))

        return self._user_settings

    def __getattr__(self, attr):

        # Any attribute must be in either required settings or defaults
        if attr not in REQUIRED_SETTINGS and attr not in self.defaults:
            raise AttributeError("Invalid DBMI setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:

            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)

        return val

    def __check_user_settings(self, user_settings):
        SETTINGS_DOC = 'https://github.com/hms-dbmi/django-dbmi-client'

        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError("The '%s' setting has been removed. Please refer to '%s' for available settings."
                                   % (setting, SETTINGS_DOC))

        if 'CLIENT' not in user_settings:
            raise AttributeError('CLIENT configuration must be set')

        # Check possible Auth0 configurations (single client, multiple clients, multiple tenants)
        if ('AUTH0_TENANT' not in user_settings or 'AUTH0_CLIENT_ID' not in user_settings) \
                and 'AUTH0_TENANTS' not in user_settings \
                and 'AUTH0_CLIENTS' not in user_settings:
            raise AttributeError('One of AUTH0_TENANT/AUTH0_CLIENT_ID, AUTH0_CLIENTS or '
                                 'AUTH0_TENANTS configurations must be set')

        # If Auth0 login enabled, ensure we've got needed Auth0 configurations
        if 'dbmi_client.login' in settings.INSTALLED_APPS \
            and ('AUTH0_TENANT' not in user_settings
                 or 'AUTH0_CLIENT_ID' not in user_settings
                 or 'AUTH0_SECRET' not in user_settings):
            raise AttributeError('AUTH0_TENANT/AUTH0_CLIENT_ID/AUTH0_SECRET must be specified for '
                                 'the "dbmi_client.login" app to function')

        # Ensure environment is set and if not prod or dev, ensure service URLs are provided
        if 'ENVIRONMENT' in user_settings and \
                user_settings['ENVIRONMENT'].lower() != 'prod' and user_settings['ENVIRONMENT'].lower() != 'dev':
            warnings.warn(
                "ENVIRONMENT is not set to production, this should be a test environment", ResourceWarning)

            missing_urls = []
            if 'AUTHN_URL' not in user_settings:
                # Check environment
                if os.environ.get('DBMI_AUTHN_URL'):
                    user_settings['AUTHN_URL'] = os.environ.get('DBMI_AUTHN_URL')
                else:
                    missing_urls.append('AUTHN_URL')

            if 'AUTHZ_URL' not in user_settings:
                # Check environment
                if os.environ.get('DBMI_AUTHZ_URL'):
                    user_settings['AUTHZ_URL'] = os.environ.get('DBMI_AUTHZ_URL')
                else:
                    missing_urls.append('AUTHZ_URL')

            if 'REG_URL' not in user_settings:
                # Check environment
                if os.environ.get('DBMI_REG_URL'):
                    user_settings['REG_URL'] = os.environ.get('DBMI_REG_URL')
                else:
                    missing_urls.append('REG_URL')
            if missing_urls:
                raise AttributeError('{} configuration(s) must be set'.format(missing_urls))

            if 'JWT_AUTHZ_NAMESPACE' not in user_settings:
                warnings.warn(
                    "JWT_AUTHZ_NAMESPACE is not set, JWT claims will not be inspected "
                    "for groups/roles/permissions", ResourceWarning)

        if 'JWT_HTTP_PREFIX' in user_settings:
            warnings.warn(
                "Changing JWT_HTTP_PREFIX breaks compatibility with DBMI services", ResourceWarning)

        if 'JWT_COOKIE_NAME' in user_settings:
            warnings.warn(
                "Changing JWT_COOKIE_NAME breaks compatibility with DBMI services", ResourceWarning)

        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')

    def __dir__(self):

        # Return defaults and user configs patched over
        attrs = dict()
        attrs.update(self.defaults)
        attrs.update(self.user_settings)

        return attrs.keys()

    @property
    def USER_MODEL_ENABLED(self):

        try:
            # Check if present in user settings
            return self.user_settings['USER_MODEL_ENABLED']

        except KeyError:

            # Else, return true if the client has added either backend classes to AUTHENTICATION_BACKENDS
            return 'dbmi_client.authn.DBMIModelAuthenticationBackend' in settings.AUTHENTICATION_BACKENDS or \
                   'dbmi_client.authn.DBMIUsersModelAuthenticationBackend' in settings.AUTHENTICATION_BACKENDS or \
                   'dbmi_client.authn.DBMIAdminModelAuthenticationBackend' in settings.AUTHENTICATION_BACKENDS

    def get_logger(self):
        """
        Returns the logger and manages whether logs propogate or not depending on user configs
        :return: logger
        """
        logger = logging.getLogger(__name__)

        # Check if disabled
        if not self.ENABLE_LOGGING:
            logger.propagate = False

        return logger


# Create the instance by which the settings should be accessed
dbmi_settings = DBMISettings(CONFIG_DEFAULTS)

# Configure logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)
logger.disabled = not dbmi_settings.ENABLE_LOGGING
logger.setLevel(dbmi_settings.LOG_LEVEL)

# Set the sentry client if Raven is used
if 'raven.contrib.django.raven_compat' in settings.INSTALLED_APPS:
    setattr(settings, 'SENTRY_CLIENT', 'dbmi_client.sentry.DBMISentryClient')


def reload_dbmi_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'DBMI_CLIENT_CONFIG':
        dbmi_settings.reload()


setting_changed.connect(reload_dbmi_settings)



from furl import furl
import requests

from rest_framework.permissions import BasePermission
from rest_framework.exceptions import PermissionDenied, NotAuthenticated

from dbmi_client.settings import dbmi_settings
from dbmi_client import authn

# Get the app logger
import logging
logger = logging.getLogger(dbmi_settings.LOGGER_NAME)

# Set keys for authz dictionary
JWT_AUTHZ_GROUPS = 'groups'
JWT_AUTHZ_ROLES = 'roles'
JWT_AUTHZ_PERMISSIONS = 'permissions'


def jwt_has_authz(claims, auth_type, item):
    """
    Inspects the JWT claims for the permission. This assumed a claims dictionary is
    namespaced under DBMIAuth.dbmi_authz_namespace and 'type' is a key to a list of items
    that 'item' will be compared to. Typically types are group, role, permission. As
    authorization claims in the JWT are not guaranteed, this will return None if those
    claims are missing, indicating the check is inconclusive and another source must
    be checked, likely DBMIAuthz.
    :param claims: The JWT claims dictionary
    :param auth_type: The type of authorization: group, role, permission
    :param item: The specific authorization to check exists for the given type
    :return: None if not defined, bool otherwise
    """
    try:
        # Call the other method with the authz claims object
        auth = claims[dbmi_settings.JWT_AUTHZ_NAMESPACE]
        return auth_has_authz(auth, auth_type, item)

    except (KeyError, IndexError, TypeError, ValueError):
        logger.debug('No authz and/or authz type ({}) in JWT claims'.format(auth_type))

    return None


def auth_has_authz(auth, auth_type, item):
    """
    Inspects the DRF auth object for the permission. This assumed a claims dictionary
    was copied from the namespaced JWT claims to the DRF auth object. See above for info
    on authorization types contained in the JWT.
    :param auth: The DRF auth object
    :param auth_type: The type of authorization: group, role, permission
    :param item: The specific authorization to check exists for the given type
    :return: None if not defined, bool otherwise
    """
    try:
        # Check groups
        for _item in auth[auth_type]:

            # Compare
            if _item == item:
                logger.debug('User has authz: {} - {}'.format(auth_type, item))
                return True

        return False

    except (KeyError, IndexError, TypeError, ValueError):
        logger.debug('No authz and/or authz type ({})'.format(auth_type))

    return None


def has_permission(request, email, item, permission):
    """
    Consults the DBMIAuthz server for authorization checks. Uses the JWT to
    authenticate the call and checks the returned permissions for the one
    specified.
    :param request: The current request containing the JWT to be checked
    :param email: The email in the JWT
    :param item: The item string to check for the permission
    :param permission: The permission to be checked for in permissions returned from DBMIAuthz
    :return: bool
    """
    url = None
    content = None
    try:
        # Build the request
        url = furl(dbmi_settings.AUTHZ_URL)
        url.path.segments.append('user_permission')
        url.path.segments.append('')
        url.query.params.add('email', email)
        url.query.params.add('item', item)
        url.query.params.add('client', dbmi_settings.CLIENT)

        # Get the JWT token depending on request type
        token = authn.get_jwt(request)
        if not token:
            return False

        # Build headers for the SciAuthZ call
        headers = {'Authorization': '{}{}'.format(dbmi_settings.JWT_HTTP_PREFIX, token),
                   'Content-Type': 'application/json'}

        # Run it
        response = requests.get(url.url, headers=headers)
        content = response.content
        response.raise_for_status()

        # Parse permissions
        for permission_result in response.json().get('results'):
            if permission_result['permission'].lower() == permission.lower():
                logger.debug('DBMIAuthZ: {} has {} on {}'.format(email, permission, dbmi_settings.CLIENT))
                return True

    except (requests.HTTPError, TypeError, KeyError):
        logger.error('SciAuthZ permission lookup failed', exc_info=True, extra={
            'request': request, 'email': email, 'permission': permission, 'url': url, 'content': content})

    return False


def is_admin(request, email):
    """
    This is just a convenience method that checks the user membership in the AUTHZ_ADMIN_GROUP, or for the
    permission specified in settings as AUTHZ_ADMIN_PERMISSION. It will check the authorization server as
    well as the JWT claims, if any.
    """
    # Check least difficult and move forward
    if jwt_has_authz(authn.get_jwt_payload(request, verify=True),
                     JWT_AUTHZ_GROUPS,
                     dbmi_settings.AUTHZ_ADMIN_GROUP):
        return True

    elif jwt_has_authz(authn.get_jwt_payload(request, verify=True),
                       JWT_AUTHZ_PERMISSIONS,
                       dbmi_settings.AUTHZ_ADMIN_PERMISSION):
        return True

    elif has_permission(request, email, dbmi_settings.CLIENT, dbmi_settings.AUTHZ_ADMIN_PERMISSION):
        return True

    return False

###################################################################
#
# Django Rest Framework (DRF) Custom Authorization
#
###################################################################


class DBMIAdminPermission(BasePermission):
    """
    Permission check for MANAGE permissions on DBMI client
    """

    def has_permission(self, request, view):

        # Get the email of the authenticated user
        if not hasattr(request, 'user'):
            logger.warning('No \'user\' attribute on request')
            raise PermissionDenied

        # Ensure claims are setup and then check them first, as it is least costly.
        if request.auth:
            if auth_has_authz(request.auth, JWT_AUTHZ_GROUPS, dbmi_settings.AUTHZ_ADMIN_GROUP):
                return True

        # Check permissions
        if has_permission(request, request.user, dbmi_settings.CLIENT, dbmi_settings.AUTHZ_ADMIN_PERMISSION):
            return True

        # Possibly store these elsewhere for records
        logger.info('{} Failed MANAGE permission for DBMI'.format(request.user))

        raise PermissionDenied


class DBMIItemPermission(BasePermission):
    """
    This permission class is meant to be inherited by clients who need to specify
    custom item strings for their permission checks. Set the item string and the permission
    and the base implementation will check the DBMI AuthZ server.
    """

    # The permission item string to check
    item = 'dbmi.item.subitem'

    # The permission itself the requesting user must have for this item
    permission = 'manage'

    def has_permission(self, request, view):

        # Get the email
        if not hasattr(request, 'user'):
            logger.warning('No \'user\' (JWT email) attribute on request')
            raise NotAuthenticated

        # Check permission server for admin permissions
        if has_permission(request, request.user, self.item, self.permission):
            return True

        raise PermissionDenied


class DBMIUserPermission(BasePermission):
    """
    Custom permission to only any authenticated user access.
    """

    def has_object_permission(self, request, view, obj):

        # Get the email
        if not hasattr(request, 'user'):
            logger.warning('No \'user\' (JWT email) attribute on request')
            raise NotAuthenticated

        return True


class DBMIOwnerPermission(BasePermission):
    """
    Object-level permission to only allow owners of an object to edit it.
    Assumes the model instance has an `email` attribute.
    """

    def has_object_permission(self, request, view, obj):

        # Get the email
        if not hasattr(request, 'user'):
            logger.warning('No \'user\' (JWT email) attribute on request')
            raise NotAuthenticated

        # Check if a key has been specified
        key = dbmi_settings.DRF_OBJECT_OWNER_KEY
        if key:

            # Ensure the attribute exists
            if not hasattr(obj, key):
                logger.error('Ownership key "{}" for object "{}" does not exist'.format(key, obj))

            else:
                logger.debug('Comparing key "{}" on "{}" for ownership'.format(key, obj))
                return getattr(obj, key) == request.user

        else:
            logger.debug('No key specified, trying attrs "email", "user" on "{}" for ownership'.format(key, obj))

            # Check email attribute
            if hasattr(obj, 'email') and obj.email == request.user:
                return True

            # Check for a user attribute
            if hasattr(obj, 'user') and obj.user == request.user:
                return True

        raise PermissionDenied


class DBMIAdminOrOwnerPermission(DBMIOwnerPermission):
    """
    Permission check for owner or MANAGE permissions on DBMI 'obj'. Owner is determined
    by comparing email in JWT with that of the email property on 'obj'
    """
    message = 'User does not have proper permission on item DBMI'

    def has_object_permission(self, request, view, obj):

        # Check ownership first
        try:
            return super(DBMIAdminOrOwnerPermission, self).has_object_permission(request, view, obj)

        except PermissionDenied:
            logger.debug('Is not owner of object, checking for admin/manage...')

        # Ensure claims are setup and then check them first, as it is least costly.
        if request.auth:
            if auth_has_authz(request.auth, JWT_AUTHZ_PERMISSIONS, dbmi_settings.AUTHZ_ADMIN_PERMISSION):
                return True

        # Lastly, check permission server for admin permissions
        if has_permission(request, request.user, dbmi_settings.CLIENT, dbmi_settings.AUTHZ_ADMIN_PERMISSION):
            return True

        # Possibly store these elsewhere for records
        logger.info('{} Failed MANAGE or owner permission for DBMI'.format(request.user))

        raise PermissionDenied
